#include "SarsaAgent.h"


using namespace std;
extern bool advicegiven;
// return a random integer in [0, maxRange)
inline int range_rand( int maxRange )
{
	return (int) ( (((double)rand()) / RAND_MAX) * maxRange );
}

inline int range_rand( int minRange, int maxRange )
{
	return range_rand( maxRange - minRange ) + minRange;
}

SarsaAgent::SarsaAgent( int numFeatures, int numActions, bool bLearn,
					   FunctionApproximator *anFA, 
					   CMAC *anFA2,
					   std::string* returnNote,
					   char *loadWeightsFile,
					   char *teacherWeightsFile
					   ): 
SMDPAgent( numFeatures, numActions )
{

	bLearning = bLearn;
	steps = 0;
	totalAdvice = 0;
	adviceCount = 0;
	exploreCount = 0;

	alpha = 0.05;
	gamma = 1.0;
	lambda = 0.00;
	epsilon = 0.01;

	epochNum = 0;
	lastAction = -1;

	FA = anFA;
	TeacherFA = anFA2;
	std::ostringstream os;

	if ( strlen( loadWeightsFile ) > 0 )
		loadWeights( loadWeightsFile );

	bFollowTeacher = false;
	
	os << strlen(teacherWeightsFile) << " " << teacherWeightsFile << " ";

	if ( strlen( teacherWeightsFile ) > 0 ){
		ifstream file;
		file.open(teacherWeightsFile);
		TeacherFA->read( &file );
		file.close();
		bFollowTeacher = true;
		os << "Read done";
	}
//bFollowTeacher = true; //TEMP!
    *returnNote += os.str(); //tempString;

}

int SarsaAgent::startEpisode( double state[], std::string* returnNote )
{
	std::string tempString;
	FA->decayTraces( 0 );
	FA->setState( state );
	std::ostringstream os;
	os << epochNum;
	for ( int a = 0; a < getNumActions(); a++ ) {
		Q[ a ] = FA->computeQ( a );
		os << " Q[" << a << "]=" << Q[a];
	}
	//*returnNote += os.str(); //tempString;

	lastAction = selectAction(returnNote);

	if (bFollowTeacher){
		lastAction = TeacherFA->getTeacherSuggestion(state, getNumActions());
		std::ostringstream os;
		os << " following with " << lastAction;
		*returnNote += os.str();
	}
	FA->updateTraces( lastAction );

	return lastAction;
}

int SarsaAgent::step( double reward, double state[], std::string* returnNote, int episodeNumber, char* adviceUnit, int numAdvice,
					 int adviceFreq, int exploreFreq, int *aCount, bool showLearning, bool cutOffEpisode,double adviceQThreshold,int Qindex )
{
	int effectiveEp;
	steps++;
	std::ostringstream os; //TEMP
	os <<endl<<" REWARD: " << reward;
		*returnNote += os.str();
	double delta = reward - Q[ lastAction ];
	FA->setState( state );
	for ( int a = 0; a < getNumActions(); a++ ) {
		Q[ a ] = FA->computeQ( a );
		os <<endl<< Q[a]; //TEMP
		*returnNote += os.str();
	}
	
	// Hook to end episode after too many steps
	if (cutOffEpisode == true){
		// move towards the enemy
		*returnNote += " Trying to die";
		return 2;

	// If this episode we are showing what we know, get our action and return it immediatly
	}else if(showLearning && ((episodeNumber % 2) != 0)){
		*returnNote += " NoLearn";
		lastAction = selectAction(returnNote);
		return lastAction;
	}

	// Get teacher advice?
	if (strcmp(adviceUnit, "step") == 0){
		if ( bFollowTeacher && (adviceCount < adviceFreq) && (totalAdvice < numAdvice) ){
			lastAction = TeacherFA->getTeacherSuggestion(state, getNumActions()); //temp
			//lastAction = TeacherFA->getTeacherSuggestion(state, getNumActions(),0,9,-1);
		/*	if (lastAction == -1)
		{
			*returnNote += " Not Important State or no student's mistake";
			lastAction = selectAction(returnNote);
		}*/
			advicegiven = true;
			std::ostringstream os;
			os << " following with " << lastAction;
			*returnNote += os.str();
			// increment advice counter
			adviceCount++;
			totalAdvice++;
			// increment external advice count
			(*aCount)++;
	
		// Use my action
		}else if ( (exploreCount < exploreFreq) || (exploreFreq == 0) ){
			lastAction = selectAction(returnNote);
			exploreCount++;

		// Done with cycle
		}else{
			adviceCount = 0;
			exploreCount = 0;
			// Decide what to do in this step

			// Get teacher advice?
			if ( bFollowTeacher && (adviceCount < adviceFreq)  && (totalAdvice < numAdvice) ){
				lastAction = TeacherFA->getTeacherSuggestion(state, getNumActions()); //temp
				//		lastAction = TeacherFA->getTeacherSuggestion(state, getNumActions(),0,9,-1);
						/*if (lastAction == -1)
		{
			*returnNote += " Not Important State or no student's mistake";
			lastAction = selectAction(returnNote);
		}*/
				advicegiven = true;
				std::ostringstream os;
				os << " following with " << lastAction;
				*returnNote += os.str();
				// increment advice counter
				adviceCount++;
				totalAdvice++;
				// increment external advice count
				(*aCount)++;
		
			// Use my action
			}else if ( (exploreCount < exploreFreq) || (exploreFreq == 0) ){
				lastAction = selectAction(returnNote);
				exploreCount++;
			}
		}
	// episode logic
	}else if((strcmp(adviceUnit, "importance") == 0)||(strcmp(adviceUnit, "mistake") == 0)){
		//std::ostringstream os;
		//os <<endl<< totalAdvice<< " YA "<<numAdvice<<endl;
		//*returnNote += os.str();
		//if ((totalAdvice < numAdvice) ){
		if ( bFollowTeacher && (totalAdvice < numAdvice) ){
			int actIntention = -1; //not known student action
		if (strcmp(adviceUnit, "mistake") == 0)
			actIntention = selectAction(returnNote);
		
		lastAction = TeacherFA->getTeacherSuggestion(state, getNumActions(),adviceQThreshold,Qindex,actIntention);
		////std::ostringstream os;
		//	//os << adviceQThreshold<< " "<<Qindex<<" "<<actIntention;
			//*returnNote += os.str();
	
		////lastAction = TeacherFA->getTeacherSuggestion(state, getNumActions());
		//	
		if (lastAction == -1)
		{
			*returnNote += " Not Important State or no student's mistake";
			lastAction = selectAction(returnNote);
		}
		else
		{
			advicegiven = true;
			*returnNote += " Advice given for an important state";
		

		//return lastAction;	
			//advicegiven = true;
			std::ostringstream os;
			os << " following with " << lastAction;
			*returnNote += os.str();
			// increment advice counter
			adviceCount++;
			totalAdvice++;
			// increment external advice count
			(*aCount)++;
		}
		// Use my action
		}else { //if ( (exploreCount < exploreFreq) || (exploreFreq == 0) )
			lastAction = selectAction(returnNote);}
			//exploreCount++;}

		// Done with cycle
		//}else{
		//	adviceCount = 0;
		//	exploreCount = 0;
		//	// Decide what to do in this step

		//	// Get teacher advice?
		//	if ( bFollowTeacher && (adviceCount < adviceFreq)  && (totalAdvice < numAdvice) ){
		//		lastAction = TeacherFA->getTeacherSuggestion(state, getNumActions()); //temp
		//
		//		advicegiven = true;
		//		std::ostringstream os;
		//		os << " following with " << lastAction;
		//		*returnNote += os.str();
		//		// increment advice counter
		//		adviceCount++;
		//		totalAdvice++;
		//		// increment external advice count
		//		(*aCount)++;
		//
		//	// Use my action
		//	}else if ( (exploreCount < exploreFreq) || (exploreFreq == 0) ){
		//		lastAction = selectAction(returnNote);
		//		exploreCount++;
		//	}
		//}

/////////////////////////////////GOOD OLD
		//if (!bFollowTeacher)
		//{
		//	*returnNote += " No loaded teacher!";
		//	lastAction = selectAction(returnNote);
		//	return lastAction;	
		//}

		//if (totalAdvice > numAdvice)
		//{
		//	*returnNote += " Not enough budget";
		//	lastAction = selectAction(returnNote);
		//	return lastAction;	
		//}

		//int actIntention = -1; //not known student action
		//if (strcmp(adviceUnit, "mistake") == 0)
		//	actIntention = selectAction(returnNote);
		//
		//lastAction = TeacherFA->getTeacherSuggestion(state, getNumActions(),adviceQThreshold,Qindex,actIntention);
		////std::ostringstream os;
		//	//os << adviceQThreshold<< " "<<Qindex<<" "<<actIntention;
		//	//*returnNote += os.str();
	
		////lastAction = TeacherFA->getTeacherSuggestion(state, getNumActions());
		//	
		//if (lastAction == -1)
		//{
		//	*returnNote += " Not Important State or no student's mistake";
		//	lastAction = selectAction(returnNote);
		//}
		//else
		//{
		//	advicegiven = true;
		//	*returnNote += " Advice given for an important state";
		//	totalAdvice++;
		//		// increment external advice count
		//	(*aCount)++;
		//}

		//return lastAction;	
		/////////////////////////////////GOOD OLD
		}
	else{
		//compute effective episode num
		if (showLearning)
			effectiveEp = episodeNumber / 2;
		else
			effectiveEp = episodeNumber;
		// Use teacher action
		if ( (adviceFreq > 0) && bFollowTeacher && ((effectiveEp % (adviceFreq + exploreFreq)) < adviceFreq) && (effectiveEp < ((exploreFreq/adviceFreq + 1)*numAdvice)) ){
			lastAction = TeacherFA->getTeacherSuggestion(state, getNumActions());// TEMP

				advicegiven = true;
			std::ostringstream os;
			os << " following with " << lastAction;
			*returnNote += os.str();
			// increment external advice count
			(*aCount)++;
	
		// Use my action
		}else {
			lastAction = selectAction(returnNote);
		}
	}

	if ( !bLearning ){
		*returnNote = *returnNote + " NoLearn";
		return lastAction;
	}

	//Learn based on lastAction
	delta += Q[ lastAction ];
	FA->updateWeights( delta, alpha );
	Q[ lastAction ] = FA->computeQ( lastAction ); // need to redo because weights changed
	FA->decayTraces( gamma * lambda );

	for ( int a = 0; a < getNumActions(); a++ ) {  //clear other than F[a]
		if ( a != lastAction ) {
			FA->clearTraces( a );
		}
	}
	FA->updateTraces( lastAction );  //replace/set traces F[a]

	return lastAction;
}

void SarsaAgent::endEpisode( double reward, int episodeNum, bool showLearning, bool cutOffEpisode ){

	epochNum++;
	if ( bLearning && lastAction != -1 ) { /* otherwise we never ran on this episode */

		/* finishing up the last episode */
		/* assuming gamma = 1  -- if not,error*/
		if ( gamma != 1.0)
			cerr << "We're assuming gamma's 1" << endl;
		if (!cutOffEpisode && !(showLearning && (episodeNum % 2 != 0))){
			double delta = reward - Q[ lastAction ];
			FA->updateWeights( delta, alpha );
		}
		ofstream log;
		log.open("steps.log");
		log << steps;
		log.close();
	}
	lastAction = -1;
}

int SarsaAgent::selectAction(std::string* returnNote)
{
	int action;
	*returnNote += " In Select";
	// Epsilon-greedy
	if (((double)rand() / RAND_MAX) < epsilon) {
		/* explore */
		action = range_rand( getNumActions() );
		*returnNote += " RandA";
	}
	else{
		action = argmaxQ(returnNote);
		*returnNote += " ArgMax";
	}
	return action;
}

bool SarsaAgent::loadWeights( char *filename )
{
	//int file = open( filename, O_RDONLY );
	ifstream file;
	file.open(filename);

	FA->read( &file );
	file.close();
	return false;
}

bool SarsaAgent::saveWeights( char *filename )
{
	//int file = open( filename, O_CREAT | O_WRONLY, 0664 );
	ofstream file;
	file.open(filename);

	FA->write( &file );
	file.close();
	return false;
}

// Returns index (action) of largest entry in Q array, breaking ties randomly 
int SarsaAgent::argmaxQ(std::string* returnNote)
{
	int bestAction = 0;
	double bestValue = Q[ bestAction ];
	int numTies = 0;
	for ( int a = bestAction + 1; a < getNumActions(); a++ ) {
		double value = Q[ a ];
		if ( value > bestValue ) {
			bestValue = value;
			bestAction = a;
		}
		else if ( value == bestValue ) {
			numTies++;
			if ( range_rand( numTies + 1 ) == 0 ) {
				bestValue = value;
				bestAction = a;
			}
		}
	}

	//*returnNote += " bestQ=";
	//std::ostringstream os;
	//os << bestValue;
	//std::string str = os.str();
	//*returnNote += str;
	return bestAction;
}

